import inspect
from typing import Generator, Any, ClassVar

import pytest

from lagom import Container, bind_to_container


class MyDep:
    loaded: ClassVar[bool] = False
    value: str

    def __init__(self, value="testing"):
        MyDep.loaded = True
        self.value = value


container = Container()


def example_function(message: str, resolved: MyDep) -> str:
    return resolved.value + message


def example_generator(message: str, resolved: MyDep) -> Generator[str, Any, Any]:
    yield resolved.value + message
    yield resolved.value + " finished"


@bind_to_container(container)
def another_example_function(message: str, resolved: MyDep) -> str:
    return resolved.value + message


def test_partial_application_can_be_applied_to_functions_with_named_args():
    partial = container.partial(example_function)
    assert partial(message=" world") == "testing world"


def test_partial_application_returns_something_that_is_considered_a_function():
    partial = container.partial(example_function)
    inspect.isfunction(partial)


def test_partial_application_can_be_applied_to_functions_with_positional_args_first():
    partial = container.partial(example_function)
    assert partial(" world") == "testing world"


def test_a_decorator_can_be_used_to_bind_as_well():
    assert another_example_function(message=" hello") == "testing hello"


def test_a_decorator_can_be_used_to_bind_and_with_positional_arguments():
    assert another_example_function(" hello") == "testing hello"


def test_container_values_can_be_overridden():
    assert (
        another_example_function(resolved=MyDep("replaced"), message=" hello")
        == "replaced hello"
    )


def test_incorrect_arguments_are_handled_well():
    with pytest.raises(TypeError) as err:
        another_example_function(not_the_message=" no")
    assert "unexpected keyword argument 'not_the_message'" in str(err.value)


def test_partial_application_can_be_applied_to_generators():
    partial = container.partial(example_generator)
    results = []
    for result in partial(message=" world"):
        results.append(result)
    assert results == ["testing world", "testing finished"]


def test_deps_are_loaded_at_call_time_not_definition_time():
    MyDep.loaded = False

    @bind_to_container(container)
    def some_random_unused_function(message: str, resolved: MyDep) -> str:
        return resolved.value + message

    assert not MyDep.loaded



import inspect
from random import random
from typing import Generator, Any, ClassVar

import pytest

from lagom import Container, bind_to_container


class SomeCache:
    loaded: ClassVar[bool] = False

    def __init__(self):
        SomeCache.loaded = True
        self._init_random = random()

    def value(self):
        return self._init_random


class MyDepOne:
    value: Any

    def __init__(self, cache: SomeCache):
        self.value = cache.value()


class MyDepTwo:
    other_value: Any

    def __init__(self, cache: SomeCache):
        self.other_value = cache.value()


def test_by_default_all_resources_are_reconstructed(container: Container):
    @bind_to_container(container)
    def example_function(dep_one: MyDepOne, dep_two: MyDepTwo):
        return {"a": dep_one.value, "b": dep_two.other_value}

    result = example_function()
    assert result["a"] != result["b"]


def test_invocation_level_singletons_can_be_defined(container: Container):
    @bind_to_container(container, shared=[SomeCache])
    def example_function_with_invocation_level_sharing(
        dep_one: MyDepOne, dep_two: MyDepTwo
    ):
        return {"a": dep_one.value, "b": dep_two.other_value}

    result = example_function_with_invocation_level_sharing()
    assert result["a"] == result["b"]


def test_invocation_level_singletons_are_not_shared_across_calls(container: Container):
    @bind_to_container(container, shared=[SomeCache])
    def example_function_with_invocation_level_sharing(
        dep_one: MyDepOne, dep_two: MyDepTwo
    ):
        return {"a": dep_one.value, "b": dep_two.other_value}

    call_one = example_function_with_invocation_level_sharing()
    call_two = example_function_with_invocation_level_sharing()

    assert call_one["a"] != call_two["a"]


def test_that_shared_types_are_lazy_loaded(container: Container):
    SomeCache.loaded = False

    @bind_to_container(container, shared=[SomeCache])
    def example_function_that_defines_but_doesnt_use_sharing():
        return "ok"

    example_function_that_defines_but_doesnt_use_sharing()
    assert not SomeCache.loaded


def test_partial_application_returns_something_that_is_considered_a_function(
    container: Container
):
    @bind_to_container(container, shared=[SomeCache])
    def example_function_with_shared():
        return "ok"

    inspect.isfunction(example_function_with_shared)



import pytest

from lagom import Construction, Container
from lagom.exceptions import DuplicateDefinition


class InitialDep:
    pass


class SomeMockForTesting(InitialDep):
    pass


class SomeMockThatDoesntEventExtend:
    pass


def test_deps_can_be_overridden_by_a_child_class(container: Container):
    container.define(InitialDep, Construction(lambda: SomeMockForTesting()))
    resolved = container.resolve(InitialDep)
    assert type(resolved) == SomeMockForTesting


def test_deps_can_be_overridden_by_anything(container: Container):
    container.define(InitialDep, Construction(lambda: SomeMockThatDoesntEventExtend()))
    resolved = container.resolve(InitialDep)
    assert type(resolved) == SomeMockThatDoesntEventExtend


def test_explicit_definitions_can_only_be_made_once(container: Container):
    container.define(InitialDep, Construction(lambda: SomeMockForTesting()))

    with pytest.raises(DuplicateDefinition):
        container.define(
            InitialDep, Construction(lambda: SomeMockThatDoesntEventExtend())
        )



from abc import ABC
from typing import List

import pytest

from lagom import Construction, Container


class MySimpleDep:
    stuff: str

    def __init__(self, stuff):
        self.stuff = stuff


class MyMoreComplicatedDep:
    stuff: str

    def __init__(self, dep: MySimpleDep):
        self.stuff = dep.stuff


class AnotherAbc(ABC):
    stuff: str = "empty"


class AnotherConcrete(AnotherAbc):
    stuff = "full"


def test_deps_can_be_referenced_by_square_brackets(container: Container):
    container[MySimpleDep] = Construction(lambda: MySimpleDep("hooray"))
    resolved = container[MySimpleDep]
    assert resolved.stuff == "hooray"


def test_construction_type_can_be_omitted(container: Container):
    container[MySimpleDep] = lambda: MySimpleDep("hooray")
    resolved = container[MySimpleDep]
    assert resolved.stuff == "hooray"


def test_singleton_type_can_be_omitted(container: Container):
    container[MySimpleDep] = MySimpleDep("hooray")
    one = container[MySimpleDep]
    two = container[MySimpleDep]
    assert one is not None
    assert one is two


def test_alias_can_be_omitted(container: Container):
    container[AnotherAbc] = AnotherConcrete
    resolved = container[AnotherAbc]
    assert type(resolved) == AnotherConcrete
    assert resolved.stuff == "full"



import pytest

from lagom import Construction, Container


class MySimpleDep:
    stuff: str

    def __init__(self, stuff):
        self.stuff = stuff


class MyMoreComplicatedDep:
    complicated_stuff: str

    def __init__(self, dep: MySimpleDep):
        self.stuff = "complicated " + dep.stuff


@pytest.fixture
def container_with_simple_dep(container: Container):
    container.define(MySimpleDep, Construction(lambda: MySimpleDep("Top stuff")))
    return container


def test_works_for_registered_types(container_with_simple_dep):
    resolved = container_with_simple_dep.resolve(MySimpleDep)
    assert resolved.stuff == "Top stuff"


def test_works_if_every_item_in_the_constructor_can_be_figured_out(
    container_with_simple_dep
):
    resolved = container_with_simple_dep.resolve(MyMoreComplicatedDep)
    assert resolved.stuff == "complicated Top stuff"


@pytest.mark.parametrize("dep", [MySimpleDep, MyMoreComplicatedDep])
def test_dependencies_are_built_each_request(container_with_simple_dep: Container, dep):
    first = container_with_simple_dep.resolve(dep)
    second = container_with_simple_dep.resolve(dep)
    assert first is not second



from dataclasses import dataclass

import pytest

from lagom import Container
from lagom.decorators import dependency_definition


@dataclass
class MyComplexDep:
    some_number: int


@dataclass
class WrapperOfSomeKind:
    inner: MyComplexDep


def test_functions_that_are_typed_can_be_used_by_a_container(container: Container):
    @dependency_definition(container)
    def my_constructor() -> MyComplexDep:
        return MyComplexDep(some_number=5)

    assert container[MyComplexDep] == MyComplexDep(some_number=5)


def test_the_functions_return_new_instances_each_time(container: Container):
    @dependency_definition(container)
    def my_constructor() -> MyComplexDep:
        return MyComplexDep(some_number=5)

    first = container[MyComplexDep]
    second = container[MyComplexDep]
    assert first is not second


def test_functions_can_be_made_into_singletons(container: Container):
    @dependency_definition(container, singleton=True)
    def my_constructor() -> MyComplexDep:
        return MyComplexDep(some_number=5)

    first = container[MyComplexDep]
    second = container[MyComplexDep]
    assert first is second


def test_definition_functions_get_an_instance_of_the_container(container: Container):
    container[MyComplexDep] = MyComplexDep(some_number=3)

    @dependency_definition(container)
    def my_constructor(c: Container) -> WrapperOfSomeKind:
        return WrapperOfSomeKind(c[MyComplexDep])

    assert container[WrapperOfSomeKind].inner == container[MyComplexDep]


def test_functions_that_are_not_typed_raise_an_error(container: Container):

    with pytest.raises(SyntaxError):

        @dependency_definition(container)
        def my_constructor():
            return MyComplexDep(some_number=5)



import pytest

from lagom.util.functional import arity


@pytest.mark.parametrize(
    "test_func,expected_arity",
    [(lambda: 0, 0), (lambda x: x, 1), (lambda x, y: x + y, 2)],
)
def test_we_can_get_arity_from_functions(test_func, expected_arity: int):
    assert arity(test_func) == expected_arity



from typing import List

import pytest

from lagom import Construction, Container


class MySimpleDep:
    stuff: str

    def __init__(self, stuff):
        self.stuff = stuff


class MyMoreComplicatedDep:
    complicated_stuff: str

    def __init__(self, dep: List[MySimpleDep]):
        self.stuff = ",".join([d.stuff for d in dep])


@pytest.fixture
def container_with_list(container: Container):
    container.define(
        List[MySimpleDep],
        Construction(lambda: [MySimpleDep("One"), MySimpleDep("Two")]),
    )
    return container


def test_works_for_list_types(container_with_list: Container):
    resolved = container_with_list.resolve(List[MySimpleDep])
    assert [x.stuff for x in resolved] == ["One", "Two"]


def test_works_for_inferred_list_types(container_with_list: Container):
    resolved = container_with_list.resolve(MyMoreComplicatedDep)
    assert resolved.stuff == "One,Two"



import inspect
from typing import List, Type

from .definitions import Singleton
from .container import Container
from .util.reflection import RETURN_ANNOTATION


def bind_to_container(container: Container, shared: List[Type] = None):
    def decorator(func):
        return container.partial(func, shared=shared)

    return decorator


def dependency_definition(container: Container, singleton: bool = False):
    def decorator(func):
        try:
            arg_spec = inspect.getfullargspec(func)
            return_type = arg_spec.annotations[RETURN_ANNOTATION]
        except KeyError:
            raise SyntaxError("Function used as a definition must have a return type")
        if singleton:
            container.define(return_type, Singleton(func))
        else:
            container.define(return_type, func)
        return func

    return decorator



import pytest

from lagom import Singleton, Construction, Container


class MyBasicDep:
    pass


class MyMoreComplicatedDep:
    def __init__(self, some_number):
        self.sum_number = some_number


class MyCompositeDep:
    def __init__(self, a: MyBasicDep, b: MyMoreComplicatedDep):
        self.a = a
        self.b = b


@pytest.fixture
def container_with_deps(container: Container):
    container.define(MyBasicDep, Singleton(MyBasicDep))
    container.define(
        MyMoreComplicatedDep, Singleton(Construction(lambda: MyMoreComplicatedDep(5)))
    )
    container.define(MyCompositeDep, Singleton(MyCompositeDep))
    return container


def test_singleton_is_only_resolved_once(container_with_deps: Container):
    first = container_with_deps.resolve(MyBasicDep)
    second = container_with_deps.resolve(MyBasicDep)
    assert first is not None
    assert first is second


def test_singleton_can_have_construction_logic(container_with_deps: Container):
    first = container_with_deps.resolve(MyMoreComplicatedDep)
    second = container_with_deps.resolve(MyMoreComplicatedDep)
    assert first.sum_number == 5
    assert first is second


def test_singleton_can_compose_other_dependencies(container_with_deps: Container):
    first = container_with_deps.resolve(MyCompositeDep)
    second = container_with_deps.resolve(MyCompositeDep)
    assert type(first.a) == MyBasicDep
    assert type(first.b) == MyMoreComplicatedDep
    assert first is second



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['lagom', 'lagom.util']

package_data = \
{'': ['*']}

setup(name='lagom',
      version='0.6.0',
      description='Lagom, a type based dependency injection container',
      author='meadsteve',
      author_email='meadsteve@gmail.com',
      url=None,
      packages=packages,
      package_data=package_data,
      python_requires='~=3.7',
     )



"""Lagom, a type based dependency injection container"""
from .version import __version__
from .definitions import Singleton, Construction, Alias
from .container import Container
from .decorators import bind_to_container
